// SimpleFacialExpression.h: interface for the SimpleFacialExpression class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SIMPLEFACIALEXPRESSION_H__67DD0330_73DE_463B_8A4E_2D67FF18DF78__INCLUDED_)
#define AFX_SIMPLEFACIALEXPRESSION_H__67DD0330_73DE_463B_8A4E_2D67FF18DF78__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "FbaAction.h"

#ifdef VISAGE_STATIC
	#define FAPLAYER_API
#else
	#ifdef FAPLAYER_EXPORTS
		#define FAPLAYER_API __declspec(dllexport)
	#else
		#define FAPLAYER_API __declspec(dllimport)
	#endif
#endif

/**
* A basic facial expression action.
* This is a simple implementation of an FbaAction, implemention the basic facial expression action. The expression
* can be one of the universal expressions - joy, sadness, anger, fear, disgust or surprise. It is set by passing
* the name of the desired expression to the constructor. The other parameter to be set is the duration time
* of the expression, in milliseconds.
*
* SimpleFacialExpression uses a simple expression dynamics scheme with attack, sustain and decay phases. 
* In the attack phase the intensity grows linearly; in the sustain phase
* it stays at the maximum; in the decay phase it falls linearly to 0. 
*
* Much more elaborate expression dynamic schemes can be implemented. 
*
* @see FbaAction
*/
class FAPLAYER_API SimpleFacialExpression : public FbaAction  
{
public:
	/**
	* Constructor.
	* Construct a simple facial expression by choosing one of the standard expressions by name, and setting its duration time.
	*
	* @param expressionName the name of the expression: joy, sadness, anger, fear, or surprise. Any other string will result in neutral expression. The expression name is not case sensitive.
	* @param expressionDuration the total duration time of the expression, in milliseconds
	* @param expressionIntensity the maximum intensity of the expression, 1.0 being the default expression intensity, higher values bringing more intense expressions
	*/
	SimpleFacialExpression(char *expressionName, long expressionDuration, float expressionIntensity);
	~SimpleFacialExpression();


	/**
	* Start the action.
	* This function is called by the FAPlayer when the playing starts. 
	*
	* @param globalTime the reference time when the action is started, in milliseconds
	*/
	void SimpleFacialExpression::start(long globalTime);

	/**
	* Get FBAPs. This function returns the Face and Body Animation Parameters (FBAPs)
	* for the time t. This function is called by the FAPlayer when the FbaAction is played in the FAPlayer.
	* This implementation implements a simple expression dynamics scheme with attack, sustain and decay phases. 
	* In the attack phase the intensity grows linearly; in the sustain phase
	* it stays at the maximum; in the decay phase it falls linearly to 0.
	*
	* @param globalTime the time for which the FBAPs are requested, in milliseconds
	* @param lastFBAPs the final FBAPs from the previous frame of animation; in this implementation we ignore this parameter
	* @param afm the Animatable Face and Body Model currently used in the player; in this implementation we ignore this parameter
	* @return the Face and Body Animation Parameters (FBAPs) for the requested time
	*/
	FBAPs * SimpleFacialExpression::getFBAPs(long globalTime, FBAPs *lastFBAPs, AFM *afm);

	/**
	* Stop playing tracks. Reset the face animation track.
	*/
	void SimpleFacialExpression::stop();

	/**
	* Return the name of the action type.
	*
	* @return the name of the action type.
	*/
	char* SimpleFacialExpression::actionTypeName() {return "SimpleFacialExpression";};

	long globalStartTime;
	float intensity;
	float maxIntensity;
	FBAPs *fbaps;
	FBAPs *expFbaps;
	long lastTimeMs;
	double attack_time; // the duration of the attack phase, in milliseconds
	double sustain_time; // the duration of the sustain phase, in milliseconds
	double decay_time; // the duration of the decay phase, in milliseconds
	int currentExpression;
	long duration;
	int playing;
};

#endif // !defined(AFX_SIMPLEFACIALEXPRESSION_H__67DD0330_73DE_463B_8A4E_2D67FF18DF78__INCLUDED_)
