/*
 * LineSocket.cpp: Base Class for LineSockets.
 *
 * Copyright (C) 2007 Almer S. Tigelaar <a.s.tigelaar@student.utwente.nl>
 * 
 * This class implements basic line-based sending/receiving operations on
 * sockets. It serves as a basis for the other LineSocket classes.
 *
 * The ideas used herein are based mostly on some Linux related socket API
 * books & some resources for the Windows (WinSock) 'flavour'. Whilst this
 * class is developed for WinSock usage, it is quite trivial to port it 
 * to GNU/Linux or other *NIX variants, since as much UNIX compatible calls
 * were used as possible.
 * ----------------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <winsock2.h>

#include "LineSocket.h"

/**
 * Creates a new LineSocket and sets up its internal buffers.
 */
LineSocket::LineSocket() {
	this->targetSocket = -1;
	this->noDelay = false;

	this->debugStream = new std::ofstream("socket_debug.txt", std::ios::app);
	//this->debugStream = &std::cerr;

	/* Initialize WinSock */
	WORD wVersionRequested = MAKEWORD(2, 2);
	WSADATA wsaData;
	if (WSAStartup(wVersionRequested, &wsaData) != 0) {
		std::cerr << "[Error] While initializing socket interface! (winsock) " << std::endl;
	}

	memset(this->buffer, '\0', sizeof(this->buffer));
	this->bufferOffset = 0;
	this->lines = new std::queue<std::string>; // Create line list
}

/**
 * Releases dynamic resources allocated during the lifetime
 * of the LineSocket.
 */
LineSocket::~LineSocket() {
	delete lines; // Release line list

	/* Clean up WinSock */
	WSACleanup();

	/* Clean up output DebugStream (but only if it is a file!) */
	if (dynamic_cast<std::fstream*>(this->debugStream)) {
		delete this->debugStream;
	}
}

/**
 * Determines whether or not a target socket has been set for reading/writing.
 *
 * @return true if a targetSocket was set, false otherwise.
 */
bool LineSocket::hasTargetSocket()
{
	return (this->targetSocket != -1);
}

/**
 * Returns the current target socket.
 *
 * @return Handle to the target socket or -1 on error
 */
SOCKET LineSocket::getTargetSocket()
{
	return this->targetSocket;
}

/**
 * Sets the target socket to use for line reading and sending.
 * NOTE: The socket is not closed / released, that is the callers responsibility.
 *
 * @param targetSocket The socket to use (or -1 to set to invalid socket).
 */
void LineSocket::setTargetSocket(SOCKET targetSocket)
{
	this->targetSocket = targetSocket;
	
	/* Automatically set the appropriate delay option on the socket */
	if (this->noDelay) {
		setsockopt(targetSocket, IPPROTO_TCP, TCP_NODELAY, (char *) this->noDelay, sizeof(this->noDelay));
	}
}

/**
 * Reads whatever data is available on the client socket and put it in the buffer.
 *
 * @return true if no errors occured (regardless if whether data was actually read), false on error.
 */
bool LineSocket::fillBuffer()
{
	if (targetSocket == -1) {
		return false;
	}

	memset(this->buffer + this->bufferOffset, '\0', sizeof(this->buffer) - this->bufferOffset);
	size_t result = recv((SOCKET) targetSocket, this->buffer + this->bufferOffset, (int) sizeof(this->buffer) - (int) this->bufferOffset, (int) 0);
	
	if (result == 0) {
		(*this->debugStream) << "[Close] Other side performed clean shutdown of connection" << std::endl;
		this->targetSocket = -1;
		this->bufferOffset = 0;
		return false;
	} else if (result == SOCKET_ERROR) {
		if (WSAGetLastError() != WSAEWOULDBLOCK) {
			(*this->debugStream) << "[Error] While reading from connection, Error Code: " << WSAGetLastError() << std::endl;
			this->targetSocket = -1;
			this->bufferOffset = 0;
		}
		return true; /* OK, no data read */
	}

	(*this->debugStream) << "[Read] " << result << " bytes from client: " << this->buffer + this->bufferOffset << std::endl;
	this->bufferOffset += result;

	return true; /* OK, data read */
}

/**
 * Parse the next line(s) available in the buffer (if any) and add these
 * to the internal list of lines.
 *
 * @return false if no lines were parsed (or on error), true if at least one line was parsed.
 */
bool LineSocket::parseBuffer()
{
	if (!fillBuffer()) {
		return false;
	}

	std::string stringBuffer = std::string(this->buffer);
	size_t newlineOffset = stringBuffer.find_first_of("\n");

	if (newlineOffset != std::string::npos) {
		/* Read the pending line and subsequent lines (if any) */
		do {
			/* Append up to the \n as new line (throws the \n itself away) */
			this->lines->push(stringBuffer.substr(0, newlineOffset));
			
			/* Shift what remains in the buffer to the front and decrease the buffer offset accordingly */
			memmove(this->buffer, this->buffer + newlineOffset + 1, sizeof(this->buffer) - newlineOffset - 1);
			this->bufferOffset -= newlineOffset + 1;

			stringBuffer = std::string(this->buffer);
			newlineOffset = stringBuffer.find_first_of("\n"); 
		} while (newlineOffset != std::string::npos);

		return true;
	}	 
	return false; /* No line read */
}

/**
 * Returns whether or not a line is available to be read.
 * Note that for each invocation of getLine() you must always invoke hasLine().
 *
 * @return true if at least one line is available for reading.
 */
bool LineSocket::hasLine()
{
	return (!this->lines->empty() || this->parseBuffer());
}

/**
 * Retrieves the next pending line on the socket (a line is always terminated by a \n, although
 * this \n is NOT included in the result of this function).
 * Note that for each invocation of getLine() you must always invoke hasLine().
 *
 * @return Next line that was pending.
 */
std::string LineSocket::getLine()
{
	if (this->lines->empty()) {
		(*this->debugStream) << "[Error] No lines pending, but getLine() invoked (forgot to check availability via hasLine()?)" << std::endl;
	}

	std::string result = this->lines->front();
	this->lines->pop();
	return result;
}

/**
 * Send a line of the socket back to the target socket.
 *
 * @param line The line to send.
 */
bool LineSocket::sendLine(std::string line)
{
	if (targetSocket == -1) {
		return false;
	}

	send(this->targetSocket, line.c_str(), (int) line.length(), 0);
	return true;
}

/**
 * Sets TCP_NO_DELAY option (this disables the Nagle algorithm).
 *
 * It is only advantageuous to set this to true if in your protocol
 * you send out little data and do not immediately receive something back.
 * (in that case it will reduce the send/receive delay between the client
 *  and the server slightly). It should not be used in any other situation!
 * Note: This option is set to 'false' by default.
 *
 * @param noDelay True to turn of packet delaying, False otherwise.
 */
void LineSocket::setNoDelay(bool noDelay)
{
	this->noDelay = noDelay;
}