/*
 * LineSocketClient.cpp: Simple Line Reading Socket Client.
 *
 * Copyright (C) 2007 Almer S. Tigelaar <a.s.tigelaar@student.utwente.nl>
 * 
 * This class simply connects to the specified port on the hostname as given
 * in the constructor. Lines can be send by using sendLine() and read by getLine(),
 * but one should always first check with hasLine() if any lines are available to be
 * read.
 *
 * The ideas used herein are based mostly on some Linux related socket API
 * books & some resources for the Windows (WinSock) 'flavour'. Whilst this
 * class is developed for WinSock usage, it is quite trivial to port it 
 * to GNU/Linux or other *NIX variants, since as much UNIX compatible calls
 * were used as necessary.
 * ----------------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <winsock2.h>

#include "LineSocketClient.h"

/**
 * Creates a new LineSocket and sets it up to listen for
 * incoming connections at the specified port.
 *
 * @param hostname The host to connect to.
 * @param port Port to listen on for connections.
 */
LineSocketClient::LineSocketClient(std::string hostname, unsigned int port) {
	struct hostent *hostInfo = gethostbyname(hostname.c_str());
	if (hostInfo == NULL) {
		(*this->debugStream) << "[Error] Problem interpreting host: " << hostname << std::endl;
		exit(1);
	}

	SOCKET socketHandle = socket(AF_INET, SOCK_STREAM, 0);
	if (socketHandle < 0) {
		(*this->debugStream) << "[Error] Cannot create socket" << std::endl;
		exit(1);
	}

	struct sockaddr_in serverAddress;
	memset(&serverAddress, '\0', sizeof(serverAddress));
	serverAddress.sin_family = hostInfo->h_addrtype;
	memcpy((char *) &serverAddress.sin_addr.s_addr, hostInfo->h_addr_list[0], hostInfo->h_length);
	serverAddress.sin_port = htons(port);
				
	if (connect(socketHandle, (struct sockaddr *) &serverAddress, sizeof(serverAddress)) < 0) {
		(*this->debugStream) << "Error: Could not connect to host" << std::endl;
	} else {
		// Make socket non-blocking
		int iMode = 1;
		ioctlsocket(socketHandle, FIONBIO, (u_long FAR*) &iMode);

		this->setTargetSocket(socketHandle);
	}
}

/**
 * Releases dynamic resources allocated during the lifetime
 * of the LineSocket.
 */
LineSocketClient::~LineSocketClient() {
	/* Clean Up Socket */
	SOCKET socketHandle = this->getTargetSocket();
	if (socketHandle != -1) {
		closesocket(socketHandle);
		this->setTargetSocket(-1);
	}
}
