/*
 * LineSocketServer.cpp: Simple Line Reading Socket Server.
 *
 * Copyright (C) 2007 Almer S. Tigelaar <a.s.tigelaar@student.utwente.nl>
 * 
 * This class creates a listening socket on the port specified in the
 * constructor. Invoking waitForClient() returns when ONE client has
 * connected to the socket (this class supports ONLY one client at a time,
 * this was done to keep it as simple and minimalistic as possible and 
 * since only 1-to-1 communication was needed for the project this class
 * is used in). After that while hasClient() is true one can call hasLine()
 * and getLine() (in that order!) to retrieve lines sent by the client 
 * one-by-one. (if hasClient() returns false, one can waitForClient() again).
 *
 * The ideas used herein are based mostly on some Linux related socket API
 * books & some resources for the Windows (WinSock) 'flavour'. Whilst this
 * class is developed for WinSock usage, it is quite trivial to port it 
 * to GNU/Linux or other *NIX variants, since as much UNIX compatible calls
 * were used as possible.
 * ----------------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <winsock2.h>

#include "LineSocketServer.h"

/**
 * Creates a new LineSocket and sets it up to listen for
 * incoming connections at the specified port.
 *
 * @param port Port to listen on for connections.
 */
LineSocketServer::LineSocketServer(unsigned int port) {
	this->socketHandle = -1;
	makeSocket(port);
}

/**
 * Releases dynamic resources allocated during the lifetime
 * of the LineSocket.
 */
LineSocketServer::~LineSocketServer() {
	/* Clean Up Sockets */
	if (this->socketHandle != -1) {
		closesocket(this->socketHandle);
		this->socketHandle = -1;
	}
	SOCKET clientHandle = this->getTargetSocket();
	if (clientHandle != -1) {
		closesocket(clientHandle);
		this->setTargetSocket(-1);
	}
}

/**
 * Make a socket, bind it to a port and set it to listening mode
 * Apart from the fact that Windows uses SOCKET (instead of 'normal' handles)
 * this is exactly similar to the UNIX approach.
 */
void LineSocketServer::makeSocket(unsigned int port)
{
	this->socketHandle = socket(AF_INET, SOCK_STREAM, 0); // Create a TCP socket object
	if (this->socketHandle < 0) {
		(*this->debugStream) << "[Error] While creating basic socket, Error Code: " << WSAGetLastError() << std::endl;
		this->socketHandle = -1;
		return;
	}

	/*
	 * Specify the bind address for the socket.
	 * This is basically just setting the port number
	 * in an elaborate way.
	 */
	struct sockaddr_in bindAddress;
	memset(&bindAddress, '\0', sizeof(bindAddress));
	bindAddress.sin_family = AF_INET;
	bindAddress.sin_port = htons(port); // htons converts the port number to network byte order
	bindAddress.sin_addr.s_addr = INADDR_ANY;

	/*
	 * Do the actual binding (i.e. tie the socket to the port)
	 */
	if (bind(this->socketHandle, (struct sockaddr *) &bindAddress, sizeof(bindAddress)) < 0) {
		(*this->debugStream) << "[Error] While binding address to socket (port in use?), Error Code: " << WSAGetLastError() << std::endl;
		this->socketHandle = -1;
	}

	/*
	 * Start listening on the port, second parameter specifies the backlog of connections
	 * to accept, since we do not want a backlog (and only accept one connection anyway), we
	 * set this to zero
	 */
	listen(this->socketHandle, 0); 
}

/**
 * This waits for a client connection, if one is obtained it returns the handle
 * to the client (basically this is the Socket descriptor for communicating with
 * the client). Apparantly there is a distinction in Windows between a File descriptor
 * and a Socket descriptor (which is not so in UNIX).
 *
 * @return true if a client has connected successfully, false on error.
 */
bool LineSocketServer::waitForClient()
{
	struct sockaddr_in clientAddress;
	int clientLen = sizeof(clientAddress);
	memset(&clientAddress, '\0', clientLen);

	SOCKET clientHandle = accept(this->socketHandle, (struct sockaddr *) &clientAddress, &clientLen);

	if (clientHandle < 0) {
		(*this->debugStream) << "[Error] While accepting client connection, Error Code: " << WSAGetLastError() << std::endl;
		return false;
	}
	(*this->debugStream) << "Got a Client!" << std::endl;

	// Make socket non-blocking
	int iMode = 1;
	ioctlsocket(clientHandle, FIONBIO, (u_long FAR*) &iMode);

	this->setTargetSocket(clientHandle);
	return true;
}

/**
 * Indicates whether or not the client link is still operational.
 * If this returns false you can call waitForClient() to accept
 * a new client connection.
 *
 * @return true if there is still a connection to a client, false otherwise.
 */
bool LineSocketServer::hasClient()
{
	return this->hasTargetSocket();
}
